"""
:summary: arundeckrun package

:license: Apache License, Version 2.0
:author: Mark LaPerriere
:contact: otupman@antillion.com
:copyright: Mark LaPerriere 2015
"""
__docformat__ = "restructuredtext en"

__VERSION__ = (0, 2, 5)
VERSION = '.'.join(map(str, __VERSION__))



"""
:summary: Python client for interacting with Rundeck API

:license: Apache License, Version 2.0
:author: Mark LaPerriere
:contact: otupman@antillion.com
:copyright: Mark LaPerriere 2015
"""
__docformat__ = "restructuredtext en"

import time
import errno
from datetime import datetime
from string import ascii_letters, digits
try:
    from string import maketrans
except ImportError:
    # python 3
    maketrans = str.maketrans

from .api import RundeckApiTolerant, RundeckApi, RundeckNode
from .connection import RundeckConnection, RundeckResponse
from .transforms import transform
from .util import child2dict, attr2dict, cull_kwargs, StringType
from .rd_exceptions import (
    RundeckServerError,
    JobNotFound,
    MissingProjectArgument,
    InvalidJobArgument,
    InvalidResponseFormat,
    InvalidJobDefinitionFormat,
    InvalidResourceSpecification,
    )
from .defaults import (
    GET,
    POST,
    DELETE,
    Status,
    DupeOption,
    UuidOption,
    JobDefFormat,
    JOB_RUN_TIMEOUT,
    JOB_RUN_INTERVAL,
    )

_JOB_ID_CHARS = ascii_letters + digits
_JOB_ID_TRANS_TAB = maketrans(_JOB_ID_CHARS, '#' * len(_JOB_ID_CHARS))
_JOB_ID_TEMPLATE = '########-####-####-####-############'
_RUNDECK_RESP_FORMATS = ('xml')  # TODO: yaml and json
_EXECUTION_COMPLETED = (Status.FAILED, Status.SUCCEEDED, Status.ABORTED)
_EXECUTION_PENDING = (Status.RUNNING,)


def is_job_id(job_id):
    """Checks if a Job ID "looks" like a UUID. It does not check if it exists as a job in Rundeck.
        And of course, a Rundeck Job ID does not have to be a "UUID". Any unique string will do
        so be prepared for false negatives if you customize your job ids.

    :Parameters:
        job_id : str
            a Rundeck Job ID

    :rtype: bool
    """
    if job_id and isinstance(job_id, StringType):
        return job_id.translate(_JOB_ID_TRANS_TAB) == _JOB_ID_TEMPLATE

    return False



class Rundeck(object):

    def __init__(self, server='localhost', protocol='http', port=4440, api_token=None, **kwargs):
        """Initialize a Rundeck API Client

        :Parameters:
            server : str
                hostname of the Rundeck server (default: localhost)
            protocol : str ('http'|'https')
                (default: 'http')
            port : int
                Rundeck server port (default: 4440)
            api_token : str
                valid Rundeck user API token

        :Keywords:
            base_path : str
                Custom base URL path for Rundeck server URLs [default: None]
            usr : str
                Rundeck user name (used in place of api_token)
            pwd : str
                Rundeck user password (used in combo with usr)
            api_version : int
                Rundeck API version
            api : RundeckApi
                an instance of a RundeckApi or subclass of RundeckApi
            connection : RundeckConnection
                an instance of a RundeckConnection or instance of a subclass of RundeckConnection
        """
        api = kwargs.pop('api', None)

        if api is None:
            self.api = RundeckApi(
                server, protocol=protocol, port=port, api_token=api_token, **kwargs)
        elif isinstance(api, RundeckApiTolerant):
            self.api = api
        else:
            raise Exception('Supplied api argument is not a valide RundeckApi: {0}'.format(api))


    @transform('system_info')
    def system_info(self, **kwargs):
        """Get Rundeck Server System Info

        :return: a dict object representing the Rundeck system information
        :rtype: dict
        """
        return self.api.system_info(**kwargs)


    def get_job_id(self, project, name=None, **kwargs):
        """Fetch a Job ID that matches the filter criterea specified
            **WARNING**: if there is more than one job that matches the specified criteria, this
                will return the first job that Rundeck server includes in the list of matches

        :Parameters:
            project : str
                the name of a project

        :Keywords:
            idlist : str | list(str, ...)
                specify a comma-separated string or a list of Job IDs to include
            groupPath : str
                specify a group or partial group path to include all jobs within that group path
                or "*" for all groups or "-" to match top level jobs only (default: "*")
            jobFilter : str
                specify a job name filter; will match any job name that contains this string
            jobExactFilter : str
                specify an exact job name to match
            groupPathExact : str
                specify an exact group path to match or "-" to match the top level jobs only

        :return: a Rundeck Job ID
        :rtype: str
        """
        if name is not None:
            kwargs['jobExactFilter'] = name

        try:
            job_list = self.get_job_ids(project, limit=1, **kwargs)
        except JobNotFound:
            raise JobNotFound('Job {0!r} not found in Project {1!r}'.format(name, project))
        else:
            return job_list[0]


    def get_job_ids(self, project, **kwargs):
        """Fetch a list of Job IDs that match the filter criterea specified

        :Parameters:
            project : str
                the name of a project

        :Keywords:
            limit : int
                limit the result set to 1 or more jobs
            idlist : str | list(str, ...)
                specify a comma-separated string or a list of Job IDs to include
            groupPath : str
                specify a group or partial group path to include all jobs within that group path
                or "*" for all groups or "-" to match top level jobs only (default: "*")
            jobFilter : str
                specify a job name filter; will match any job name that contains this string
            jobExactFilter : str
                specify an exact job name to match
            groupPathExact : str
                specify an exact group path to match or "-" to match the top level jobs only

        :return: a list of Job IDs
        :rtype: list
        """
        limit = kwargs.pop('limit', None)
        resp = self.list_jobs(project, **kwargs)
        jobs = resp

        job_ids = []
        if len(jobs) > 0:
            job_ids = [job['id'] for job in jobs]
        else:
            raise JobNotFound('No jobs in Project {0!r} matching criteria'.format(project))

        return job_ids[:limit]  # if limit is None, this will return the whole shebang


    @transform('jobs')
    def list_jobs(self, project, **kwargs):
        """Fetch a listing of jobs for a project

        :Parameters:
            project : str
                the name of a project

        :Keywords:
            limit : int
                limit the result set to 1 or more jobs
            idlist : str | list(str, ...)
                specify a comma-separated string or a list of Job IDs to include
            groupPath : str
                specify a group or partial group path to include all jobs within that group path
                or "*" for all groups or "-" to match top level jobs only (default: "*")
            jobFilter : str
                specify a job name filter; will match any job name that contains this string
            jobExactFilter : str
                specify an exact job name to match
            groupPathExact : str
                specify an exact group path to match or "-" to match the top level jobs only

        :return: a list of Jobs
        :rtype: list(dict, ...)
        """
        jobs = self.api.jobs(project, **kwargs)
        jobs.raise_for_error()
        return jobs


    def run_job(self, job_id, **kwargs):
        """Wraps job_run method and implements a blocking mechanism to wait for the job to
            complete (within reason, i.e. timeout and interval)

            See Rundeck.job_run docstring for additional keyword args

        :Parameters:
            job_id : str
                Rundeck Job ID

        :Keywords:
            timeout : int | float
                how many seconds to wait for a completed status (default: 60)
            interval : int | float
                how many seconds to sleep between polling cycles (default: 3)
            argString : str | dict
                argument string to pass to job - if str, will be passed as-is
                else if dict will be converted to compatible string
            loglevel : str('DEBUG', 'VERBOSE', 'INFO', 'WARN', 'ERROR')
                logging level (default: 'INFO')
            asUser : str
                user to run the job as
            exclude-precedence : bool
                set the exclusion precedence (default True)
            hostname : str
                hostname inclusion filter
            tags : str
                tags inclusion filter
            os-name : str
                os-name inclusion filter
            os-family : str
                os-family inclusion filter
            os-arch : str
                os-arch inclusion filter
            os-version : str
                os-version inclusion filter
            name : str
                name inclusion filter
            exlude-hostname : str
                hostname exclusion filter
            exlude-tags : str
                tags exclusion filter
            exlude-os-name : str
                os-name exclusion filter
            exlude-os-family : str
                os-family exclusion filter
            exlude-os-arch : str
                os-arch exclusion filter
            exlude-os-version : str
                os-version exclusion filter
            exlude-name : str
                name exclusion filter

        :return: Details about the Job Execution
        :rtype: dict
        """
        timeout = kwargs.pop('timeout', JOB_RUN_TIMEOUT)
        interval = kwargs.pop('interval', JOB_RUN_INTERVAL)

        execution = self._run_job(job_id, **kwargs)

        exec_id = execution['id']
        start = time.time()
        duration = 0

        while (duration < timeout):

            execution = self.execution_status(exec_id)

            try:
                exec_status = execution['status']
            except AttributeError:
                # for some reason, we don't always immediately get an execution back from Rundeck
                #    loop once before we let the execption bubble up
                if duration == 0:
                    continue

            if exec_status in _EXECUTION_COMPLETED:
                break

            time.sleep(interval)
            duration = time.time() - start

        return execution


    @transform('execution')
    def _run_job(self, job_id, **kwargs):
        """Kick off a Rundeck Job

        :Parameters:
            job_id : str
                Rundeck Job ID

        :Keywords:
            argString : str | dict
                argument string to pass to job - if str, will be passed as-is
                else if dict will be converted to compatible string
            loglevel : str('DEBUG', 'VERBOSE', 'INFO', 'WARN', 'ERROR')
                logging level (default: 'INFO')
            asUser : str
                user to run the job as
            exclude-precedence : bool
                set the exclusion precedence (default True)
            hostname : str
                hostname inclusion filter
            tags : str
                tags inclusion filter
            os-name : str
                os-name inclusion filter
            os-family : str
                os-family inclusion filter
            os-arch : str
                os-arch inclusion filter
            os-version : str
                os-version inclusion filter
            name : str
                name inclusion filter
            exlude-hostname : str
                hostname exclusion filter
            exlude-tags : str
                tags exclusion filter
            exlude-os-name : str
                os-name exclusion filter
            exlude-os-family : str
                os-family exclusion filter
            exlude-os-arch : str
                os-arch exclusion filter
            exlude-os-version : str
                os-version exclusion filter
            exlude-name : str
                name exclusion filter

        :return: Details about the Job Execution
        :rtype: dict
        """
        return self.api.job_run(job_id, **kwargs)


    def jobs_export(self, project, **kwargs):
        """Export a the job definitions for a project in XML or YAML format

        :Parameters:
            name : str
                name of the project

        :Keywords:
            fmt : str ('xml'|'yaml')
                format of the definition string (default: 'xml')
            idlist : list(str, ...)
                a list of job ids to return
            groupPath : str
                a group path, partial group path or the special top level only
                char '-'
            jobFilter: str
                find job names that include this string

        :return: a job definition
        :rtype: str
        """
        return self.api.jobs_export(project, **kwargs).text


    @transform('job_import_status')
    def import_job(self, definition, **kwargs):
        """Import a job definition string in XML or YAML format

        :Parameters:
            definition : str
                a string representing a job definition

        :Keywords:
            fmt : str ('xml'|'yaml')
                format of the definition string (default: 'xml')
            dupeOption : str ('skip'|'create'|'update')
                a value to indicate the behavior when importing jobs which already exist
                (default: 'create')
            project : str
                specify the project that all job definitions should be imported to otherwise all
                job definitions must define a project
            uuidOption : str ('preserve'|'remove')
                preserve or remove UUIDs in imported jobs - preserve may fail if a UUID already
                exists

        :return: the results of the job import
        :rtype: dict
        """
        return self.api.jobs_import(definition, **kwargs)


    def import_job_file(self, file_path, **kwargs):
        """Read in the contents of a job definition file for import

        :Parameters:
            file_path : str
                the path to a readable job definition file

        :Keywords:
            file_format : str ('xml'|'yaml')
                if not specified it will be derived from the file extension (default: 'xml')

        :raise IOError: raised if job definition file can not be found or is not readable

        :return: a dict object representing a set of Rundeck status messages
        :rtype: dict
        """
        fmt = kwargs.pop('file_format', None)
        if fmt is None:
            fmt = JobDefFormat.XML
            fmt_specified = False
        else:
            fmt_specified = True

        definition = open(file_path, 'r').read()

        if not fmt_specified:
            fmt = os.path.splitext(file_name.strip())[1][1:].lower()

        if fmt not in JobDefFormat.values:
            raise InvalidJobDefinitionFormat('Invalid Job definition format: {0}'.format(fmt))

        return self.import_job(definition, fmt=fmt, **kwargs)


    def export_job(self, job_id, **kwargs):
        """Export a job definition in XML or YAML format

        :Parameters:
            job_id : str
                Rundeck Job ID

        :Keywords:
            fmt : str
                the format of the response one of JobDefFormat.values (default: 'xml')

        :return: a Job definition
        :rtype: str
        """
        return self.api.job(job_id, **kwargs)


    def delete_job(self, job_id, **kwargs):
        """Delete a job

        :Parameters:
            job_id : str
                Rundeck Job ID

        :return: success
        :rtype: bool
        """
        result = self.api.delete_job(job_id, **kwargs)
        # api version 11 wil respond with a 204 No Content; older version use the result xml node
        if self.api_version >= 11:
            if result.response.status_code == 204:
                return True
            else:
                return False
        else:
            rd_msg = RundeckResponse(result)
            return rd_msg.success


    def delete_jobs(self, idlist, **kwargs):
        """Bulk Job delete

        :Parameters:
            idlist : str | list(str, ...)
                a list of job ids or a string of comma seperated job ids to delete

        :return: results of the bulk delete with details
        :rtype: dict

        Example Response::

            {
            'requestCount': 3,
            'allsuccessful': False,
            'succeeded': {
                'id': '1234-123-123-12345',
                'message': 'success',
                },
            'failed': {
                'id': '9876-986-987-98765',
                'message': 'Job ID 9876-986-987-98765 does not exist',
                },
            }

        """
        # while we're waiting for https://github.com/dtolabs/rundeck/issues/588 to get resolved,
        #   we'll just use iterate over the list of ids and call delete_job - potentially REALLY
        #   painfully slow for large lists
        if isinstance(idlist, StringType):
            idlist = idlist.split(',')

        results = []
        for id in idlist:
            result = None
            try:
                result = self.delete_job(id)
            except RundeckServerError as exc:
                result = exc.rundeck_response
            results.append(result)

        return results


    @transform('executions')
    def list_job_executions(self, job_id, **kwargs):
        """Get a list of executions of a Job

        :Parameters:
            job : str
                A Job ID

        :Keywords:
            status : str
                one of Status.values
            max : int
                maximum number of results to include in response (default: 20)
            offset : int
                offset for result set (default: 0)

        :return: a list of Job Executions
        :rtype: list(dict, ...)
        """
        return self.api.job_executions(job_id, **kwargs)


    @transform('executions')
    def list_running_executions(self, project='*', **kwargs):
        """Retrieve running executions

        :Parameters:
            project : str
                the name of a project (use "*" for all projects - API v9+ required)

        :return: a list of Job Executions
        :rtype: list(dict, ...)
        """
        return self.api.executions_running(project, **kwargs)


    @transform('execution')
    def execution_status(self, execution_id, **kwargs):
        """Get the status of an execution

        :Parameters:
            execution_id : int
                Rundeck Job Execution ID

        :return: an Execution
        :rtype: dict
        """
        return self.api.execution(execution_id, **kwargs)


    @transform('executions')
    def query_executions(self, project, **kwargs):
        """Execution query

        :Parameters:
            project : str
                name of the project

        :Keywords:
            statusFilter : str
                one of Status.values
            abortedbyFilter : str
                user name that aborted the execution
            userFilter : str
                user name that initiated the execution
            recentFilter | str
                Use a simple text format to filter executions that completed within a period of
                time; the format is "XY" where 'X' is an integer and 'Y' is one of:

                    * `h`:hour
                    * `d`:day
                    * `w`:week
                    * `m`:month
                    * `y`:year

                So a value of "2w" would return executions that completed within the last two weeks

            begin : int | str
                either a unix millisecond timestamp or a W3C dateTime "yyyy-MM-ddTHH:mm:ssZ"
            end : int | str
                either a unix millisecond timestamp or a W3C dateTime "yyyy-MM-ddTHH:mm:ssZ"
            adhoc : bool
                includes adhoc executions
            jobIdListFilter : str | list
                one or more job ids to include
            excludeJobIdListFilter : str | list
                one or more job ids to exclude
            jobListFilter : str | list
                one or more full job group/name to include
            excludeJobListFilter : str | list
                one or more full job group/name to include
            groupPath : str
                a group or partial group path to include, special "-" setting matches top level
                jobs only
            groupPathExact : str
                an exact group path to include, special "-" setting matches top level jobs only
            excludeGroupPath : str
                a group or partial group path to exclude, special "-" setting matches top level
                jobs only
            excludeGroupPathExact : str
                an exact group path to exclude, special "-" setting matches top level jobs only
            jobExactFilter : str
                an exact job name
            exludeJobExactFilter : str
                an exact job name to exclude
            max : int
                maximum number of results to include in response (default: 20)
            offset : int
                offset for result set (default: 0)

        :return: a list of Job Executions
        :rtype: list(dict, ...)
        """
        return self.api.executions(project, **kwargs)


    @transform('execution_output')
    def _execution_output_json(self, execution_id, **kwargs):
        return self.api.execution_output(execution_id, **kwargs)


    def get_execution_output(self, execution_id, **kwargs):
        """Get the output for an execution in various formats

        :Parameters:
            execution_id : str
                Rundeck Job Execution ID

        :Keywords:
            fmt : str
                the format of the response one of ExecutionOutputFormat.values (default: 'json')
            raw : bool
                return the results of the Execution output request unmodified (default: False)
            offset : int
                byte offset to read from in the file, 0 indicates the beginning
            lastlines : int
                number of lines to retrieve from the end of the available output, overrides offset
            lastmod : int
                a unix millisecond timestamp; return output data received after this timestamp
            maxlines : int
                maximum number of lines to retrieve forward from the specified offset

        :return: Execution output
        :rtype: str | dict | RundeckResponse
        """
        raw = kwargs.pop('raw', None)
        fmt = kwargs.pop('fmt', None)
        if fmt is None and raw is None:
            fmt = 'json'
            raw = False

        if fmt is None:
            # default raw is in xml format
            fmt = 'xml'

        if raw or fmt == 'text':
            return self.api.execution_output(execution_id, fmt=fmt, **kwargs).text
        elif fmt == 'json':
            return self._execution_output_json(execution_id, fmt=fmt, **kwargs)
        elif fmt == 'xml':
            return self.api.execution_output(execution_id, fmt=fmt, parse_response=True, **kwargs)


    @transform('execution_abort')
    def abort_execution(self, execution_id, **kwargs):
        """Abort a running Job Execution

        :Parameters:
            execution_id : str
                Rundeck Job Execution ID

        :Keywords:
            asUser : str
                specifies a username identifying the user who aborted the execution

        :return: abort status information
        :rtype: dict
        """
        return self.api.execution_abort(execution_id, **kwargs)


    @transform('run_execution')
    def run_adhoc_command(self, project, command, **kwargs):
        """Run a command

        :Parameters:
            project : str
                name of the project
            command : str
                the shell command string to run


        :Keywords:
            nodeThreadcount : int
                the number of threads to use
            nodeKeepgoing : bool
                if True, continue executing on other nodes even if some fail
            asUser : str
                specifies a username identifying the user who ran the command; requires runAs
                permission
            hostname : str
                hostname inclusion filter
            tags : str
                tags inclusion filter
            os-name : str
                os-name inclusion filter
            os-family : str
                os-family inclusion filter
            os-arch : str
                os-arch inclusion filter
            os-version : str
                os-version inclusion filter
            name : str
                name inclusion filter
            exlude-hostname : str
                hostname exclusion filter
            exlude-tags : str
                tags exclusion filter
            exlude-os-name : str
                os-name exclusion filter
            exlude-os-family : str
                os-family exclusion filter
            exlude-os-arch : str
                os-arch exclusion filter
            exlude-os-version : str
                os-version exclusion filter
            exlude-name : str
                name exclusion filter

        :return: Execution ID
        :rtype: int
        """
        return self.api.run_command(project, command, **kwargs)


    @transform('run_execution')
    def run_adhoc_script(self, project, scriptFile, **kwargs):
        """Run a script downloaded from a URL

        :Parameters:
            project : str
                name of the project
            scriptFile : str
                a string containing the script file content


        :Keywords:
            argString : str | dict
                argument string to pass to job - if str, will be passed as-is
                else if dict will be converted to compatible string
            nodeThreadcount : int
                the number of threads to use
            nodeKeepgoing : bool
                if True, continue executing on other nodes even if some fail
            asUser : str
                specifies a username identifying the user who ran the command; requires runAs
                permission
            scriptInterpreter : str
                a command to use to run the script (requires API version 8 or higher)
            interpreterArgsQuoted : bool
                if True the script file and arguments will be quoted as the last argument to the
                scriptInterpreter (requires API version 8 or higher)
            hostname : str
                hostname inclusion filter
            tags : str
                tags inclusion filter
            os-name : str
                os-name inclusion filter
            os-family : str
                os-family inclusion filter
            os-arch : str
                os-arch inclusion filter
            os-version : str
                os-version inclusion filter
            name : str
                name inclusion filter
            exlude-hostname : str
                hostname exclusion filter
            exlude-tags : str
                tags exclusion filter
            exlude-os-name : str
                os-name exclusion filter
            exlude-os-family : str
                os-family exclusion filter
            exlude-os-arch : str
                os-arch exclusion filter
            exlude-os-version : str
                os-version exclusion filter
            exlude-name : str
                name exclusion filter

        :return: Execution ID
        :rtype: int
        """
        return self.api.run_script(project, scriptFile, **kwargs)


    @transform('run_execution')
    def run_adhoc_url(self, project, scriptUrl, **kwargs):
        """Run a script downloaded from a URL

        :Parameters:
            project : str
                name of the project
            scriptUrl : str
                a URL referencing a script to download and run


        :Keywords:
            argString : str | dict
                argument string to pass to job - if str, will be passed as-is
                else if dict will be converted to compatible string
            nodeThreadcount : int
                the number of threads to use
            nodeKeepgoing : bool
                if True, continue executing on other nodes even if some fail
            asUser : str
                specifies a username identifying the user who ran the command; requires runAs
                permission
            scriptInterpreter : str
                a command to use to run the script (requires API version 8 or higher)
            interpreterArgsQuoted : bool
                if True the script file and arguments will be quoted as the last argument to the
                scriptInterpreter (requires API version 8 or higher)
            hostname : str
                hostname inclusion filter
            tags : str
                tags inclusion filter
            os-name : str
                os-name inclusion filter
            os-family : str
                os-family inclusion filter
            os-arch : str
                os-arch inclusion filter
            os-version : str
                os-version inclusion filter
            name : str
                name inclusion filter
            exlude-hostname : str
                hostname exclusion filter
            exlude-tags : str
                tags exclusion filter
            exlude-os-name : str
                os-name exclusion filter
            exlude-os-family : str
                os-family exclusion filter
            exlude-os-arch : str
                os-arch exclusion filter
            exlude-os-version : str
                os-version exclusion filter
            exlude-name : str
                name exclusion filter

        :return: Execution ID
        :rtype: int
        """
        return self.api.run_url(project, scriptUrl, **kwargs)


    @transform('projects')
    def list_projects(self, **kwargs):
        """Get a list of projects

        :return: a list of Rundeck projects
        :rtype: list(dict, ...)
        """
        return self.api.projects(GET, **kwargs)


    @transform('project')
    def get_project(self, project, **kwargs):
        """Fetch a project details

        :Parameters:
            project : str
                the name of a project

        :return: detailed information about a project
        :rtype: dict
        """
        return self.api.project(project, **kwargs)


    @transform('project')
    def create_project(self, project, **kwargs):
        """Create a project
        Requires API version >11

        :Parameters:
            project : str
                name of the project

        :Keywords:
            config : dict
                a dictionary of key/value pairs for the project config

        :return: detailed information about a project
        :rtype: dict
        """
        return self.api.projects(POST, project=project, **kwargs)


    @transform('project_resources')
    def _project_resources(self, project, **kwargs):
        """Transforms a Rundeck.project_resources response
        """
        return self.api.project_resources(project, **kwargs)


    def list_project_resources(self, project, **kwargs):
        """Retrieve the list of resources for a project. If `fmt` is unspecified, a python
        dictionary will be returned

        :Parameters:
            project : str
                name of the project

        :Keywords:
            fmt : str
                the format of the response, either 'python' (dict), 'xml' or 'yaml'
                    (default: 'python')
            hostname : str
                hostname inclusion filter
            tags : str
                tags inclusion filter
            os-name : str
                os-name inclusion filter
            os-family : str
                os-family inclusion filter
            os-arch : str
                os-arch inclusion filter
            os-version : str
                os-version inclusion filter
            name : str
                name inclusion filter
            exlude-hostname : str
                hostname exclusion filter
            exlude-tags : str
                tags exclusion filter
            exlude-os-name : str
                os-name exclusion filter
            exlude-os-family : str
                os-family exclusion filter
            exlude-os-arch : str
                os-arch exclusion filter
            exlude-os-version : str
                os-version exclusion filter
            exlude-name : str
                name exclusion filter

        :return: A list of resources or a string representing the requested resources in the
            requested format
        :rtype: list({'name': str, 'hostname': str, 'username': str)) | str
        """
        fmt = kwargs.pop('fmt', 'python')

        if fmt is 'python':
            return self._project_resources(project, quiet=True, **kwargs)
        else:
            return self.api.project_resources(project, fmt=fmt, parse_response=False, **kwargs).text


    @transform('success_message')
    def update_project_resources(self, project, nodes, **kwargs):
        """Update the resources for a project

        :Parameters:
            project : str
                name of the project
            nodes : list
                a list of nodes in the form of a three tuple (`(name, hostname, username)`) or
                a dictionary at least the following keys 'name', 'hostname' and 'username'

        :return: success message
        :rtype: dict
        """
        if isinstance(nodes, tuple):
            nodes = [nodes]
        elif isinstance(nodes, dict):
            nodes = [nodes]
        elif not isinstance(nodes, list):
            raise InvalidResourceSpecification(
                "'nodes' must be a tuple or a dict or a list of tuples or dicts")

        required_keys = ('name', 'hostname', 'username')
        rundeck_nodes = []
        for node in nodes:
            if isinstance(node, dict) and set(node.keys()).issuperset(required_keys):
                rundeck_nodes.append(
                    RundeckNode(
                        node.pop('name'),
                        node.pop('hostname'),
                        node.pop('username'),
                        **node
                        )
                    )
            elif isinstance(node, tuple) and len(node) == 3:
                rundeck_nodes.append(RundeckNode(*node))

        if len(rundeck_nodes) > 0:
            return self.api.project_resources_update(project, rundeck_nodes, **kwargs)
        else:
            raise InvalidResourceSpecification('No valid nodes provided')

    @transform('project_archive_import')
    def import_project_archive(self, project_name, archive_zip_data, **kwargs):
        """Wraps `Rundeck API PUT /api/14/project/[PROJECT]/import <http://rundeck.org/docs/api/index.html#project-archive-import>`_

        :Parameters:
            project : str
                name of the project
            archive_zip_data : ?
                Zip file containing a project archive.

        :Keywords:
            jobUuidOption : str ('preserve'|'remove')
                Option declaring how duplicate Job UUIDs should be handled. If preserve (default) then
                imported job UUIDs will not be modified, and may conflict with jobs in other projects.
                If remove then all job UUIDs will be removed before importing.
            importExecutions : bool
                 import all executions and logs from the archive (default: true)
            importConfig : bool
                 import project configurationfrom the archive (default: true)
            importACL : bool
                 import all of the ACL policies from the archive (default: true)

        :return: A :class:`~.rundeck.connection.RundeckResponse`
        :rtype: :class:`~.rundeck.connection.RundeckResponse`
        """
        return self.api.project_archive_import(project_name, archive_zip_data, **kwargs)


    @transform('success_message')
    def refresh_project_resources(self, project, providerURL=None, **kwargs):
        """Refresh the resources for a project via its Resource Model Provider URL

        :Parameters:
            project : str
                name of the project
            providerURL : str
                Specify the Resource Model Provider URL to refresh the resources from; otherwise
                the configured provider URL in the `project.properties` file will be used

        :return: success message
        :rtype: dict
        """
        return self.api.project_resources_refresh(project, providerURL=providerURL, **kwargs)


    @transform('events')
    def get_project_history(self, project, **kwargs):
        """List history events for a project

        :Parameters:
            project : str
                name of the project

        :Keywords:
            jobIdFilter : str
                include event for a job ID
            reportIdFilter : str
                include events for an event name
            userFilter : str
                include events created by user
            statFilter : str
                one of Status.values
            jobListFilter : str | list
                one or more full job group/name to include
            excludeJobListFilter : str | list
                one or more full job group/name to include
            recentFilter | str
                Use a simple text format to filter executions that completed within a period of
                time; the format is "XY" where 'X' is an integer and 'Y' is one of:

                    * `h`:hour
                    * `d`:day
                    * `w`:week
                    * `m`:month
                    * `y`:year

                So a value of "2w" would return executions that completed within the last two weeks

            begin : int | str
                either a unix millisecond timestamp or a W3C dateTime "yyyy-MM-ddTHH:mm:ssZ"
            end : int | str
                either a unix millisecond timestamp or a W3C dateTime "yyyy-MM-ddTHH:mm:ssZ"
            max : int
                maximum number of results to include in response (default: 20)
            offset : int
                offset for result set (default: 0)


        :return: A RundeckResponse
        :rtype: RundeckResponse
        """
        return self.api.history(project, **kwargs)



"""
:summary: Utility functions for the package

:license: Apache License, Version 2.0
:author: Mark LaPerriere
:contact: otupman@antillion.com
:copyright: Mark LaPerriere 2015
"""
__docformat__ = "restructuredtext en"


def child2dict(el):
    """Turns an ElementTree.Element's children into a dict using the node names as dict keys and
    and the node text as dict values

    :Parameters:
        el : ElementTree.Element

    :return: a dictionary of element key(tag name)/value(node text) pairs
    :rtype: dict
    """
    return {c.tag: c.text for c in el}


def attr2dict(el):
    """Turns an elements attrib dict into... wait for it... a dict. Yea, it's silly to me too.
    But, according to the ElementTree docs, using the Element.attrib attribute directly
    is not recommended - don't look at me - I just work here.

    :Parameters:
        el : ElementTree.Element

    :return: a dictionary of element attrib key/value pairs
    :rtype: dict
    """
    return {k: v for k, v in el.items()}


def node2dict(el):
    """Combines both the attr2dict and child2dict functions
    """
    return dict(list(attr2dict(el).items()) + list(child2dict(el).items()))


def cull_kwargs(api_keys, kwargs):
    """Strips out the api_params from kwargs based on the list of api_keys
    !! modifies kwargs inline

    :Parameters:
        api_keys : list | set | tuple
            an iterable representing the keys of the key value pairs to pull out of kwargs
        kwargs : dict
            a dictionary of kwargs

    :return: a dictionary the API params
    :rtype: dict
    """
    return {k: kwargs.pop(k) for k in api_keys if k in kwargs}


def dict2argstring(argString):
    """Converts an argString dict into a string otherwise returns the string unchanged

    :Parameters:
        argString : str | dict
            argument string to pass to job - if str, will be passed as-is else if dict will be
            converted to compatible string

    :return: an argString
    :rtype: str
    """
    if isinstance(argString, dict):
        return ' '.join(['-' + str(k) + ' ' + str(v) for k, v in argString.items()])
    else:
        return argString


try:
    if isinstance('', basestring):
        pass
except NameError:
    # python 3
    StringType = type('')
else:
    # python 2
    StringType = basestring



"""
:summary: Test rundeck.client.Rundeck

:license: Apache License, Version 2.0
:author: Mark LaPerriere
:contact: otupman@antillion.com
:copyright: Mark LaPerriere 2015

:requires: nose
"""
__docformat__ = "restructuredtext en"

from nose.tools import raises


@raises(NotImplementedError)
def test_is_job_id():
    # is_job_id(job_id)
    raise NotImplementedError('Test not yet implemented')


@raises(NotImplementedError)
def test_system_info():
    # system_info(self, **kwargs)
    raise NotImplementedError('Test not yet implemented')


@raises(NotImplementedError)
def test_get_job_id():
    # get_job_id(self, project, name=None, **kwargs)
    raise NotImplementedError('Test not yet implemented')


@raises(NotImplementedError)
def test_get_job_ids():
    # get_job_ids(self, project, **kwargs)
    raise NotImplementedError('Test not yet implemented')


@raises(NotImplementedError)
def test_list_jobs():
    # list_jobs(self, project, **kwargs)
    raise NotImplementedError('Test not yet implemented')


@raises(NotImplementedError)
def test_run_job():
    # run_job(self, project, job_name, **kwargs)
    raise NotImplementedError('Test not yet implemented')


@raises(NotImplementedError)
def test_jobs_export():
    # jobs_export(self, project, **kwargs)
    raise NotImplementedError('Test not yet implemented')


@raises(NotImplementedError)
def test_import_job():
    # import_job(self, definition, **kwargs)
    raise NotImplementedError('Test not yet implemented')


@raises(NotImplementedError)
def test_import_job_file():
    # import_job_file(self, file_path, **kwargs)
    raise NotImplementedError('Test not yet implemented')


@raises(NotImplementedError)
def test_export_job():
    # export_job(self, job_id, **kwargs)
    raise NotImplementedError('Test not yet implemented')


@raises(NotImplementedError)
def test_delete_job():
    # delete_job(self, job_id, **kwargs)
    raise NotImplementedError('Test not yet implemented')


@raises(NotImplementedError)
def test_delete_jobs():
    # delete_jobs(self, idlist, **kwargs)
    raise NotImplementedError('Test not yet implemented')


@raises(NotImplementedError)
def test_list_job_executions():
    # list_job_executions(self, job_id, **kwargs)
    raise NotImplementedError('Test not yet implemented')


@raises(NotImplementedError)
def test_list_running_executions():
    # list_running_executions(self, project='*', **kwargs)
    raise NotImplementedError('Test not yet implemented')


@raises(NotImplementedError)
def test_execution_status():
    # execution_status(self, execution_id, **kwargs)
    raise NotImplementedError('Test not yet implemented')


@raises(NotImplementedError)
def test_query_executions():
    # query_executions(self, project, **kwargs)
    raise NotImplementedError('Test not yet implemented')


@raises(NotImplementedError)
def test_get_execution_output():
    # get_execution_output(self, execution_id, **kwargs)
    raise NotImplementedError('Test not yet implemented')


@raises(NotImplementedError)
def test_abort_execution():
    # abort_execution(self, execution_id, **kwargs)
    raise NotImplementedError('Test not yet implemented')


@raises(NotImplementedError)
def test_run_adhoc_command():
    # run_adhoc_command(self, project, command, **kwargs)
    raise NotImplementedError('Test not yet implemented')


@raises(NotImplementedError)
def test_run_adhoc_script():
    # run_adhoc_script(self, project, scriptFile, **kwargs)
    raise NotImplementedError('Test not yet implemented')


@raises(NotImplementedError)
def test_run_adhoc_url():
    # run_adhoc_url(self, project, scriptUrl, **kwargs)
    raise NotImplementedError('Test not yet implemented')


@raises(NotImplementedError)
def test_list_projects():
    # list_projects(self, **kwargs)
    raise NotImplementedError('Test not yet implemented')


@raises(NotImplementedError)
def test_get_project():
    # get_project(self, project, **kwargs)
    raise NotImplementedError('Test not yet implemented')


@raises(NotImplementedError)
def test_list_project_resources():
    # list_project_resources(self, project, **kwargs)
    raise NotImplementedError('Test not yet implemented')


@raises(NotImplementedError)
def test_update_project_resources():
    # update_project_resources(self, project, nodes, **kwargs)
    raise NotImplementedError('Test not yet implemented')


@raises(NotImplementedError)
def test_refresh_project_resources():
    # refresh_project_resources(self, project, providerURL=None, **kwargs)
    raise NotImplementedError('Test not yet implemented')


@raises(NotImplementedError)
def test_get_project_history():
    # get_project_history(self, project, **kwargs)
    raise NotImplementedError('Test not yet implemented')




"""
:summary: Direct interaction with Rundeck server API

:license: Apache License, Version 2.0
:author: Mark LaPerriere
:contact: otupman@antillion.com
:copyright: Mark LaPerriere 2015
"""
__docformat__ = "restructuredtext en"

from functools import partial
from xml.sax.saxutils import quoteattr
try:
    from urllib import quote as urlquote
except ImportError:
    # python 3
    from urllib.parse import quote as urlquote

from .connection import RundeckConnectionTolerant, RundeckConnection
from .util import cull_kwargs, dict2argstring, StringType
from .rd_exceptions import (
    InvalidResponseFormat,
    InvalidJobDefinitionFormat,
    InvalidDupeOption,
    InvalidUuidOption,
    HTTPError
    )
from .defaults import (
    GET,
    POST,
    PUT,
    DELETE,
    DupeOption,
    UuidOption,
    JobDefFormat,
    ExecutionOutputFormat,
    RUNDECK_API_VERSION,
    )

def api_version_check(api_version, required_version):
    """Raises a NotImplementedError if the api_version of the connection isn't sufficient
    """
    if api_version < required_version:
        raise NotImplementedError('Call requires API version {0} or higher'.format(required_version))


class RundeckNode(object):
    """Represents a Rundeck node; mainly for serializing to XML

    :IVariables:
        name : str
            name of the node
        hostname : str
            hostname of the node
        username : str
            username used for the remote connection
        description : str
            node description
        osArch : str
            the node's operating system architecture
        osFamily : str
            the node's operating system system family
        osName : str
            the node's operating system name
        tags : list(str, ...)
            a list of filtering tags
        editUrl : str
            URL to an external resource model editor service
        remoteUrl : str
            URL to an external resource model service
        attributes : dict
            a dictionary of name/value pairs to be used as node attributes
    """

    def __init__(self, name, hostname, username, **kwargs):
        """Initialize a RundeckNode instance

        :Parameters:
            name : str
                name of the node
            hostname : str
                hostname of the node
            username : str
                username used for the remote connection

        :Keywords:
            description : str
                node description
            osArch : str
                the node's operating system architecture
            osFamily : str
                the node's operating system system family
            osName : str
                the node's operating system name
            tags : list(str, ...)
                a list of filtering tags
            editUrl : str
                URL to an external resource model editor service
            remoteUrl : str
                URL to an external resource model service
            attributes : dict
                a dictionary of name/value pairs to be used as node attributes
        """
        self.name = name
        self.hostname = hostname
        self.username = username

        self.description = kwargs.get('description', None)
        self.osArch = kwargs.get('osArch', None)
        self.osFamily = kwargs.get('osFamily', None)
        self.osName = kwargs.get('osName', None)
        self.tags = kwargs.get('tags', None)
        self.editUrl = kwargs.get('editUrl', None)
        self.remoteUrl = kwargs.get('remoteUrl', None)
        self.attributes = kwargs.get('attributes', None)


    def serialize(self):
        """serializes the instance to XML

        :rtype: str
        :return: an XML string
        """
        node_attr_keys = (
            'name',
            'hostname',
            'username',
            'description',
            'osArch',
            'osFamily',
            'osName',
            'editUrl',
            'remoteUrl',
        )

        data = {k: getattr(self, k)
                    for k in node_attr_keys if getattr(self, k, None) is not None}

        if self.tags is not None and hasattr(self.tags, '__iter__'):
            data['tags'] = ','.join(self.tags)
        elif isinstance(self.tags, StringType):
            data['tags'] = self.tags

        node_xml_attrs = ' '.join(['{0}={1}'.format(k, quoteattr(v)) for k, v in data.items()])

        node_attributes = ''
        if self.attributes is not None and isinstance(self.attributes, dict):
            node_attributes = ''.join(['<attribute name="{0}" value="{1}" />'.format(k, v)
                                for k, v in self.attributes.items()])

        return '<node {0}>{1}</node>'.format(node_xml_attrs, node_attributes)

    @property
    def xml(self):
        return self.serialize()



class RundeckApiTolerant(object):
    """As close to the Rundeck API as possible. The "Tolerant" class does not throw exceptions
    when HTTP status codes are returned. Probably don't want/need to use this.

    :IVariables:
        connection : :class:`~rundeck.connection.RundeckConnection`
            a :class:`~rundeck.connection.RundeckConnection` instance
    """

    def __init__(self, server='localhost', protocol='http', port=4440, api_token=None, **kwargs):
        """Initialize a Rundeck API instance

        :Parameters:
            server : str
                hostname of the Rundeck server (default: localhost)
            protocol : str ('http'|'https')
                (default: 'http')
            port : int
                Rundeck server port (default: 4440)
            api_token : str
                valid Rundeck user API token

        :Keywords:
            base_path : str
                Custom base URL path for Rundeck server URLs [default: None]
            usr : str
                Rundeck user name (used in place of api_token)
            pwd : str
                Rundeck user password (used in combo with usr)
            api_version : int
                Rundeck API version
            connection : :class:`~rundeck.connection.RundeckConnection`
                an instance of a :class:`~rundeck.connection.RundeckConnection` or instance of a
                subclass of :class:`~rundeck.connection.RundeckConnection`
        """
        connection = kwargs.pop('connection', None)

        if connection is None:
            self.connection = RundeckConnection(
                server=server, protocol=protocol, port=port, api_token=api_token, **kwargs)
        elif isinstance(connection, RundeckConnectionTolerant):
            self.connection = connection
        else:
            raise Exception('Supplied connection argument is not a valid RundeckConnection: {0}'.format(connection))

        self.requires_version = partial(api_version_check, self.connection.api_version)


    def _exec(self, method, url, params=None, data=None, parse_response=True, **kwargs):
        """ Executes a request to Rundeck via the RundeckConnection

        :Parameters:
            method : str
                either :class:`~rundeck.defaults.GET` or :class:`~rundeck.defaults.POST`
            url : str
                Rundeck API endpoint URL
            params : dict
                dict of query string params
            data : dict
                dict of POST data

        :Keywords:
            \*\* : \*
                all remaining keyword arguments will be passed on to RundeckConnection._exec

        :return: A :class:`~.rundeck.connection.RundeckResponse`
        :rtype: :class:`~.rundeck.connection.RundeckResponse`
        """
        return self.connection.call(
            method, url, params=params, data=data, parse_response=parse_response, **kwargs)


    def system_info(self, **kwargs):
        """Wraps `Rundeck API GET /system/info <http://rundeck.org/docs/api/index.html#system-info>`_

        :return: A :class:`~.rundeck.connection.RundeckResponse`
        :rtype: :class:`~.rundeck.connection.RundeckResponse`
        """
        return self._exec(GET, 'system/info', **kwargs)


    def jobs(self, project, **kwargs):
        """Wraps `Rundeck API GET /jobs <http://rundeck.org/docs/api/index.html#listing-jobs>`_

        :Parameters:
            project : str
                the name of a project

        :Keywords:
            idlist : str | list(str, ...)
                specify a comma-separated string or a list of Job IDs to include
            groupPath : str
                specify a group or partial group path to include all jobs within that group path
                or "*" for all groups or "-" to match top level jobs only (default: "*")
            jobFilter : str
                specify a job name filter; will match any job name that contains this string
            jobExactFilter : str
                specify an exact job name to match
            groupPathExact : str
                specify an exact group path to match or "-" to match the top level jobs only

        :return: A :class:`~.rundeck.connection.RundeckResponse`
        :rtype: :class:`~.rundeck.connection.RundeckResponse`
        """
        params = cull_kwargs(
            ('idlist', 'groupPath', 'jobFilter', 'jobExactFilter', 'groupPathExact'), kwargs)

        if 'jobExactFilter' in params or 'groupPathExact' in params:
            self.requires_version(2)

        params['project'] = project

        return self._exec(GET, 'jobs', params=params, **kwargs)


    def project_jobs(self, project, **kwargs):
        """ Simulates `Rundeck API GET /project/[NAME]/jobs <http://rundeck.org/docs/api/index.html#listing-jobs-for-a-project>`_
            **Note**: Can't find any difference between this and /jobs, so this method is just a
                synonym for the `jobs` method

        :Parameters:
            project : str
                name of the project

        :Keywords:
            idlist : list(str, ...)
                a list of job ids to return
            groupPath : str
                a group path, partial group path or the special top level only
                char '-'
            jobFilter: str
                find job names that include this string
            jobExactFilter : str
                a specific job name to return
            groupPathExact : str
                a exact group path to match or the special top level only char
                '-'

        :return: A :class:`~.rundeck.connection.RundeckResponse`
        :rtype: :class:`~.rundeck.connection.RundeckResponse`
        """
        return self.jobs(project, **kwargs)


    def job_run(self, job_id, **kwargs):
        """Wraps `Rundeck API GET /job/[ID]/run <http://rundeck.org/docs/api/index.html#running-a-job>`_

        :Parameters:
            job_id : str
                Rundeck Job ID

        :Keywords:
            argString : str | dict
                argument string to pass to job - if str, will be passed as-is
                else if dict will be converted to compatible string
            loglevel : str('DEBUG', 'VERBOSE', 'INFO', 'WARN', 'ERROR')
                logging level (default: 'INFO')
            asUser : str
                user to run the job as
            exclude-precedence : bool
                set the exclusion precedence (default True)
            hostname : str
                hostname inclusion filter
            tags : str
                tags inclusion filter
            os-name : str
                os-name inclusion filter
            os-family : str
                os-family inclusion filter
            os-arch : str
                os-arch inclusion filter
            os-version : str
                os-version inclusion filter
            name : str
                name inclusion filter
            exlude-hostname : str
                hostname exclusion filter
            exlude-tags : str
                tags exclusion filter
            exlude-os-name : str
                os-name exclusion filter
            exlude-os-family : str
                os-family exclusion filter
            exlude-os-arch : str
                os-arch exclusion filter
            exlude-os-version : str
                os-version exclusion filter
            exlude-name : str
                name exclusion filter

        :return: A :class:`~.rundeck.connection.RundeckResponse`
        :rtype: :class:`~.rundeck.connection.RundeckResponse`
        """
        params = cull_kwargs(('argString', 'loglevel', 'asUser', 'exclude-precedence', \
            'hostname', 'tags', 'os-name', 'os-family', 'os-arch', 'os-version', 'name', \
            'exlude-hostname', 'exlude-tags', 'exlude-os-name', 'exlude-os-family', \
            'exlude-os-arch', 'exlude-os-version', 'exlude-name'), kwargs)

        argString = params.get('argString', None)
        if argString is not None:
            params['argString'] = dict2argstring(argString)

        return self._exec(GET, 'job/{0}/run'.format(job_id), params=params, **kwargs)


    def jobs_export(self, project, **kwargs):
        """Wraps `Rundeck API GET /jobs/export <http://rundeck.org/docs/api/index.html#exporting-jobs>`_

        :Parameters:
            name : str
                name of the project

        :Keywords:
            fmt : str ('xml'|'yaml')
                format of the definition string (default: 'xml')
            idlist : list(str, ...)
                a list of job ids to return
            groupPath : str
                a group path, partial group path or the special top level only
                char '-'
            jobFilter: str
                find job names that include this string

        :return: A Requests response
        :rtype: requests.models.Response
        """
        params = cull_kwargs(('fmt', 'idlist', 'groupPath', 'jobFilter'), kwargs)
        if 'fmt' in params:
            params['format'] = params.pop('fmt')
        params['project'] = project

        return self._exec(GET, 'jobs/export', params=params, parse_response=False, **kwargs)


    def jobs_import(self, definition, **kwargs):
        """Wraps `Rundeck API POST /jobs/import <http://rundeck.org/docs/api/index.html#importing-jobs>`_

        :Parameters:
            definition : str
                a string representing a job definition

        :Keywords:
            fmt : str ('xml'|'yaml')
                format of the definition string (default: 'xml')
            dupeOption : str ('skip'|'create'|'update')
                a value to indicate the behavior when importing jobs which already exist
                (default: 'create')
            project : str
                specify the project that all job definitions should be imported to otherwise all
                job definitions must define a project
            uuidOption : str ('preserve'|'remove')
                preserve or remove UUIDs in imported jobs - preserve may fail if a UUID already
                exists

        :return: A :class:`~.rundeck.connection.RundeckResponse`
        :rtype: :class:`~.rundeck.connection.RundeckResponse`
        """
        data = cull_kwargs(('fmt', 'dupeOption', 'project', 'uuidOption'), kwargs)
        data['xmlBatch'] = definition
        if 'fmt' in data:
            data['format'] = data.pop('fmt')

        return self._exec(POST, 'jobs/import', data=data, **kwargs)


    def job(self, job_id, **kwargs):
        """Wraps `Rundeck API GET /job/[ID] <http://rundeck.org/docs/api/index.html#getting-a-job-definition>`_

        :Parameters:
            job_id : str
                Rundeck Job ID

        :Keywords:
            fmt : str
                the format of the response one of :class:`~rundeck.defaults.JobDefFormat` ``values``
                (default: 'xml')

        :return: A Requests response
        :rtype: requests.models.Response
        """
        params = cull_kwargs(('fmt',), kwargs)

        if 'fmt' in params:
            params['format'] = params.pop('fmt')

        return self._exec(GET, 'job/{0}'.format(job_id), params=params, parse_response=False, **kwargs)


    def delete_job(self, job_id, **kwargs):
        """Wraps `Rundeck API DELETE /job/[ID] <http://rundeck.org/docs/api/index.html#deleting-a-job-definition>`_

        :Parameters:
            job_id : str
                Rundeck Job ID

        :return: A Requests response
        :rtype: requests.models.Response
        """
        return self._exec(DELETE, 'job/{0}'.format(job_id), parse_response=False, **kwargs)


    def jobs_delete(self, idlist, **kwargs):
        """Wraps `Rundeck API POST /jobs/delete <http://rundeck.org/docs/api/index.html#importing-jobs>`_

        :Parameters:
            ids : str | list(str, ...)
                a list of job ids or a string of comma seperated job ids to delete

        :return: A :class:`~.rundeck.connection.RundeckResponse`
        :rtype: :class:`~.rundeck.connection.RundeckResponse`
        """
        if not isinstance(idlist, StringType) and hasattr(idlist, '__iter__'):
            idlist = ','.join(idlist)

        data = {
            'idlist': idlist,
            }

        try:
            return self._exec(POST, 'jobs/delete', data=data, **kwargs)
        except Exception as exc:
            # TODO: what on earth did I do there?!?! need to fix this
            raise


    def job_executions(self, job_id, **kwargs):
        """Wraps `Rundeck API GET /job/[ID]/executions <http://rundeck.org/docs/api/index.html#getting-executions-for-a-job>`_

        :Parameters:
            job_id : str
                a Job ID

        :Keywords:
            status : str
                one of :class:`~rundeck.defaults.Status` ``values``
            max : int
                maximum number of results to include in response (default: 20)
            offset : int
                offset for result set (default: 0)

        :return: A :class:`~.rundeck.connection.RundeckResponse`
        :rtype: :class:`~.rundeck.connection.RundeckResponse`
        """
        params = cull_kwargs(('status', 'max', 'offset'), kwargs)
        return self._exec(GET, 'job/{0}/executions'.format(job_id), params=params, **kwargs)


    def executions_running(self, project, **kwargs):
        """Wraps `Rundeck API GET /executions/running <http://rundeck.org/docs/api/index.html#listing-running-executions>`_

        :Parameters:
            project : str
                the name of a project

        :return: A :class:`~.rundeck.connection.RundeckResponse`
        :rtype: :class:`~.rundeck.connection.RundeckResponse`
        """
        params = {'project': project}
        return self._exec(GET, 'executions/running', params=params, **kwargs)


    def execution(self, execution_id, **kwargs):
        """Wraps `Rundeck API GET /execution/[ID] <http://rundeck.org/docs/api/index.html#getting-execution-info>`_

        :Parameters:
            execution_id : str
                Rundeck Job Execution ID

        :return: A :class:`~.rundeck.connection.RundeckResponse`
        :rtype: :class:`~.rundeck.connection.RundeckResponse`
        """
        return self._exec(GET, 'execution/{0}'.format(execution_id), **kwargs)


    def executions(self, project, **kwargs):
        """Wraps `Rundeck API GET /executions <http://rundeck.org/docs/api/index.html#getting-execution-info>`_

        :Parameters:
            project : str
                name of the project

        :Keywords:
            statusFilter : str
                one of :class:`~rundeck.defaults.Status` ``values``
            abortedbyFilter : str
                user name that aborted the execution
            userFilter : str
                user name that initiated the execution
            recentFilter | str
                Use a simple text format to filter executions that completed within a period of
                time; the format is "XY" where 'X' is an integer and 'Y' is one of:

                    * `h`:hour
                    * `d`:day
                    * `w`:week
                    * `m`:month
                    * `y`:year

                So a value of "2w" would return executions that completed within the last two weeks

            begin : int | str
                either a unix millisecond timestamp or a W3C dateTime "yyyy-MM-ddTHH:mm:ssZ"
            end : int | str
                either a unix millisecond timestamp or a W3C dateTime "yyyy-MM-ddTHH:mm:ssZ"
            adhoc : bool
                includes adhoc executions
            jobIdListFilter : str | list
                one or more job ids to include
            excludeJobIdListFilter : str | list
                one or more job ids to exclude
            jobListFilter : str | list
                one or more full job group/name to include
            excludeJobListFilter : str | list
                one or more full job group/name to include
            groupPath : str
                a group or partial group path to include, special "-" setting matches top level
                jobs only
            groupPathExact : str
                an exact group path to include, special "-" setting matches top level jobs only
            excludeGroupPath : str
                a group or partial group path to exclude, special "-" setting matches top level
                jobs only
            excludeGroupPathExact : str
                an exact group path to exclude, special "-" setting matches top level jobs only
            jobExactFilter : str
                an exact job name
            exludeJobExactFilter : str
                an exact job name to exclude
            max : int
                maximum number of results to include in response (default: 20)
            offset : int
                offset for result set (default: 0)

        :return: A :class:`~.rundeck.connection.RundeckResponse`
        :rtype: :class:`~.rundeck.connection.RundeckResponse`
        """
        self.requires_version(5)

        params = cull_kwargs(('statusFilter', 'abortedbyFilter', 'userFilter', 'recentFilter', \
            'begin', 'end', 'adhoc', 'jobIdListFilter', 'excludeJobIdListFilter', \
            'jobListFilter', 'excludeJobListFilter', 'groupPath', 'groupPathExact', \
            'excludeGroupPath', 'excludeGroupPathExact', 'jobExactFilter', \
            'exludeJobExactFilter', 'max', 'offset'), kwargs)
        params['project'] = project

        return self._exec(GET, 'executions', params=params, **kwargs)


    def execution_output(self, execution_id, **kwargs):
        """Wraps `Rundeck API GET /execution/[ID]/output <http://rundeck.org/docs/api/index.html#execution-output>`_

        :Parameters:
            execution_id : str
                Rundeck Job Execution ID

        :Keywords:
            fmt : str
                the format of the response one of :class:`~rundeck.defaults.ExecutionOutputFormat`
                ``values`` (default: 'text')
            offset : int
                byte offset to read from in the file, 0 indicates the beginning
            lastlines : int
                number of lines to retrieve from the end of the available output, overrides offset
            lastmod : int
                a unix millisecond timestamp; return output data received after this timestamp
            maxlines : int
                maximum number of lines to retrieve forward from the specified offset

        :return: A Requests response
        :rtype: requests.models.Response
        """
        params = cull_kwargs(('fmt', 'offset', 'lastlines', 'lastmod', 'maxlines'), kwargs)
        if 'fmt' in params:
            params['format'] = params.pop('fmt')

        parse_response = kwargs.pop('parse_response', False)

        return self._exec(GET, 'execution/{0}/output'.format(execution_id), params=params, parse_response=parse_response, **kwargs)


    def execution_abort(self, execution_id, **kwargs):
        """Wraps `Rundeck API GET /execution/[ID]/output <http://rundeck.org/docs/api/index.html#execution-output>`_

        :Parameters:
            execution_id : str
                Rundeck Job Execution ID

        :Keywords:
            asUser : str
                specifies a username identifying the user who aborted the execution

        :return: A Requests response
        :rtype: requests.models.Response
        """
        params = cull_kwargs(('asUser',), kwargs)
        return self._exec(GET, 'execution/{0}/abort'.format(execution_id), params=params, **kwargs)


    def run_command(self, project, command, **kwargs):
        """Wraps `Rundeck API GET /run/command <http://rundeck.org/docs/api/index.html#running-adhoc-commands>`_

        :Parameters:
            project : str
                name of the project
            command : str
                the shell command string to run


        :Keywords:
            nodeThreadcount : int
                the number of threads to use
            nodeKeepgoing : bool
                if True, continue executing on other nodes even if some fail
            asUser : str
                specifies a username identifying the user who ran the command; requires runAs
                permission
            hostname : str
                hostname inclusion filter
            tags : str
                tags inclusion filter
            os-name : str
                os-name inclusion filter
            os-family : str
                os-family inclusion filter
            os-arch : str
                os-arch inclusion filter
            os-version : str
                os-version inclusion filter
            name : str
                name inclusion filter
            exlude-hostname : str
                hostname exclusion filter
            exlude-tags : str
                tags exclusion filter
            exlude-os-name : str
                os-name exclusion filter
            exlude-os-family : str
                os-family exclusion filter
            exlude-os-arch : str
                os-arch exclusion filter
            exlude-os-version : str
                os-version exclusion filter
            exlude-name : str
                name exclusion filter

        :return: A :class:`~.rundeck.connection.RundeckResponse`
        :rtype: :class:`~.rundeck.connection.RundeckResponse`
        """
        params = cull_kwargs(('nodeThreadcount', 'nodeKeepgoing', 'asUser', 'hostname', 'tags', \
            'os-name', 'os-family', 'os-arch', 'os-version', 'name', 'exlude-hostname', \
            'exlude-tags', 'exlude-os-name', 'exlude-os-family', 'exlude-os-arch', \
            'exlude-os-version', 'exlude-name'), kwargs)

        params['project'] = project
        params['exec'] = command

        return self._exec(GET, 'run/command', params=params, **kwargs)


    def run_script(self, project, scriptFile, **kwargs):
        """Wraps `Rundeck API POST /run/script <http://rundeck.org/docs/api/index.html#running-adhoc-scripts>`_

        :Parameters:
            project : str
                name of the project
            scriptFile : str
                a string containing the script file content


        :Keywords:
            argString : str | dict
                argument string to pass to job - if str, will be passed as-is
                else if dict will be converted to compatible string
            nodeThreadcount : int
                the number of threads to use
            nodeKeepgoing : bool
                if True, continue executing on other nodes even if some fail
            asUser : str
                specifies a username identifying the user who ran the command; requires runAs
                permission
            scriptInterpreter : str
                a command to use to run the script (requires API version 8 or higher)
            interpreterArgsQuoted : bool
                if True the script file and arguments will be quoted as the last argument to the
                scriptInterpreter (requires API version 8 or higher)
            hostname : str
                hostname inclusion filter
            tags : str
                tags inclusion filter
            os-name : str
                os-name inclusion filter
            os-family : str
                os-family inclusion filter
            os-arch : str
                os-arch inclusion filter
            os-version : str
                os-version inclusion filter
            name : str
                name inclusion filter
            exlude-hostname : str
                hostname exclusion filter
            exlude-tags : str
                tags exclusion filter
            exlude-os-name : str
                os-name exclusion filter
            exlude-os-family : str
                os-family exclusion filter
            exlude-os-arch : str
                os-arch exclusion filter
            exlude-os-version : str
                os-version exclusion filter
            exlude-name : str
                name exclusion filter

        :return: A :class:`~.rundeck.connection.RundeckResponse`
        :rtype: :class:`~.rundeck.connection.RundeckResponse`
        """
        params = cull_kwargs(('argString', 'nodeThreadcount', 'nodeKeepgoing', 'asUser', \
            'scriptInterpreter', 'interpreterArgsQuoted', 'hostname', 'tags', 'os-name', \
            'os-family', 'os-arch', 'os-version', 'name', 'exlude-hostname', 'exlude-tags', \
            'exlude-os-name', 'exlude-os-family', 'exlude-os-arch', 'exlude-os-version', \
            'exlude-name'), kwargs)

        params['project'] = project
        files = {'scriptFile': scriptFile}

        if 'scriptInterpreter' in params or 'interpreterArgsQuoted' in params:
            self.requires_version(8)

        argString = params.get('argString', None)
        if argString is not None:
            params['argString'] = dict2argstring(argString)


        return self._exec(POST, 'run/script', params=params, files=files, **kwargs)


    def run_url(self, project, scriptURL, **kwargs):
        """Wraps `Rundeck API POST /run/url <http://rundeck.org/docs/api/index.html#running-adhoc-script-urls>`_
        Requires API version >4

        :Parameters:
            project : str
                name of the project
            scriptURL : str
                a URL referencing a script to download and run


        :Keywords:
            argString : str | dict
                argument string to pass to job - if str, will be passed as-is
                else if dict will be converted to compatible string
            nodeThreadcount : int
                the number of threads to use
            nodeKeepgoing : bool
                if True, continue executing on other nodes even if some fail
            asUser : str
                specifies a username identifying the user who ran the command; requires runAs
                permission
            scriptInterpreter : str
                a command to use to run the script (requires API version 8 or higher)
            interpreterArgsQuoted : bool
                if True the script file and arguments will be quoted as the last argument to the
                scriptInterpreter (requires API version 8 or higher)
            hostname : str
                hostname inclusion filter
            tags : str
                tags inclusion filter
            os-name : str
                os-name inclusion filter
            os-family : str
                os-family inclusion filter
            os-arch : str
                os-arch inclusion filter
            os-version : str
                os-version inclusion filter
            name : str
                name inclusion filter
            exlude-hostname : str
                hostname exclusion filter
            exlude-tags : str
                tags exclusion filter
            exlude-os-name : str
                os-name exclusion filter
            exlude-os-family : str
                os-family exclusion filter
            exlude-os-arch : str
                os-arch exclusion filter
            exlude-os-version : str
                os-version exclusion filter
            exlude-name : str
                name exclusion filter

        :return: A :class:`~.rundeck.connection.RundeckResponse`
        :rtype: :class:`~.rundeck.connection.RundeckResponse`
        """
        self.requires_version(4)

        data = cull_kwargs(('argString', 'nodeThreadcount', 'nodeKeepgoing', 'asUser', \
            'scriptInterpreter', 'interpreterArgsQuoted', 'hostname', 'tags', 'os-name', \
            'os-family', 'os-arch', 'os-version', 'name', 'exlude-hostname', 'exlude-tags', \
            'exlude-os-name', 'exlude-os-family', 'exlude-os-arch', 'exlude-os-version', \
            'exlude-name'), kwargs)

        data['project'] = project
        data['scriptURL'] = scriptURL

        if 'scriptInterpreter' in data or 'interpreterArgsQuoted' in data:
            self.requires_version(8)

        argString = data.get('argString', None)
        if argString is not None:
            data['argString'] = dict2argstring(argString)

        return self._exec(POST, 'run/url', data=data, **kwargs)


    def _post_projects(self, project, **kwargs):
        """Wraps `Rundeck API POST /projects <http://rundeck.org/docs/api/index.html#project-creation>`_
        Requires API version >11

        :Parameters:
            project : str
                name of the project

        :Keywords:
            config : dict
                a dictionary of key/value pairs for the project config

        :return: A :class:`~.rundeck.connection.RundeckResponse`
        :rtype: :class:`~.rundeck.connection.RundeckResponse`
        """
        self.requires_version(11)

        config = kwargs.pop('config', None)

        prop_tmpl = '<property key="{0}" value="{1}" />'
        config_tmpl =  '  <config>\n' + \
                       '    {0}\n' + \
                       '  </config>\n'
        project_tmpl = '<project>\n' + \
                       '  <name>{0}</name>\n' + \
                       '{1}</project>'

        if config is not None:
            props_xml = '    \n'.join([prop_tmpl.format(k, v) for k, v in config.items()])
            config_xml = config_tmpl.format(props_xml)
        else:
            config_xml = ''

        xml = project_tmpl.format(project, config_xml)
        print(xml)

        headers = {'Content-Type': 'application/xml'}

        return self._exec(POST, 'projects', data=xml, headers=headers, **kwargs)


    def _get_projects(self, **kwargs):
        """Wraps `Rundeck API GET /projects <http://rundeck.org/docs/api/index.html#listing-projects>`_

        :return: A :class:`~.rundeck.connection.RundeckResponse`
        :rtype: :class:`~.rundeck.connection.RundeckResponse`
        """
        return self._exec(GET, 'projects', **kwargs)


    def projects(self, method=GET, **kwargs):
        """An interface to the Rundeck `projects` API endpoint support

        :Parameters:
            method : str
                GET will retrieve a project, POST will create a project

        :return: A :class:`~.rundeck.connection.RundeckResponse`
        :rtype: :class:`~.rundeck.connection.RundeckResponse`
        """
        if method == GET:
            return self._get_projects(**kwargs)
        elif method == POST:
            return self._post_projects(**kwargs)


    def project(self, project, **kwargs):
        """Wraps `Rundeck API /project/[NAME] <http://rundeck.org/docs/api/index.html#getting-project-info>`_

        :Parameters:
            project : str
                name of Project

        :Keywords:
            create : bool
                Create the project if it is not found (requires API version >11)
                [default: True for API version >11 else False]

        :return: A :class:`~.rundeck.connection.RundeckResponse``
        :rtype: :class:`~.rundeck.connection.RundeckResponse`
        """
        # check if project creation is desired and apply defaults
        create = kwargs.pop('create', None)
        if create is None:
            if self.connection.api_version >= 11:
                create = True
            else:
                create = False
        elif create == True:
            self.requires_version(11)

        rd_url = 'project/{0}'.format(urlquote(project))

        # seed project var (seems cleaner than alternatives)
        project = None
        try:
            project = self._exec(GET, rd_url, **kwargs)
        except HTTPError as exc:
            if create:
                project = self._exec(POST, rd_url, **kwargs)
            else:
                raise

        return project


    def project_resources(self, project, **kwargs):
        """Wraps `Rundeck API GET /project/[NAME]/resources <http://rundeck.org/docs/api/index.html#updating-and-listing-resources-for-a-project>`_

        :Parameters:
            project : str
                name of the project

        :Keywords:
            fmt : str
                the format of the response one of :class:`~rundeck.defaults.ExecutionOutputFormat`
                ``values`` (default: 'text')
            hostname : str
                hostname inclusion filter
            tags : str
                tags inclusion filter
            os-name : str
                os-name inclusion filter
            os-family : str
                os-family inclusion filter
            os-arch : str
                os-arch inclusion filter
            os-version : str
                os-version inclusion filter
            name : str
                name inclusion filter
            exlude-hostname : str
                hostname exclusion filter
            exlude-tags : str
                tags exclusion filter
            exlude-os-name : str
                os-name exclusion filter
            exlude-os-family : str
                os-family exclusion filter
            exlude-os-arch : str
                os-arch exclusion filter
            exlude-os-version : str
                os-version exclusion filter
            exlude-name : str
                name exclusion filter

        :return: A :class:`~.rundeck.connection.RundeckResponse`
        :rtype: :class:`~.rundeck.connection.RundeckResponse`
        """
        self.requires_version(2)

        params = cull_kwargs(('fmt', 'scriptInterpreter', 'interpreterArgsQuoted', 'hostname', \
            'tags', 'os-name', 'os-family', 'os-arch', 'os-version', 'name', 'exlude-hostname', \
            'exlude-tags', 'exlude-os-name', 'exlude-os-family', 'exlude-os-arch', \
            'exlude-os-version', 'exlude-name'), kwargs)

        if 'fmt' in params:
            params['format'] = params.pop('fmt')

        return self._exec(GET, 'project/{0}/resources'.format(urlquote(project)), params=params, **kwargs)


    def project_resources_update(self, project, nodes, **kwargs):
        """Wraps `Rundeck API POST /project/[NAME]/resources <http://rundeck.org/docs/api/index.html#updating-and-listing-resources-for-a-project>`_

        :Parameters:
            project : str
                name of the project
            nodes : list(RundeckNode, ...)
                a list of RundeckNode objects

        :return: A :class:`~.rundeck.connection.RundeckResponse`
        :rtype: :class:`~.rundeck.connection.RundeckResponse`
        """
        headers = {'Content-Type': 'text/xml'}

        data = '<nodes>{0}</nodes>'.format('\n'.join([node.xml for node in nodes]))

        return self._exec(POST, 'project/{0}/resources'.format(urlquote(project)), data=data, headers=headers, **kwargs)


    def project_resources_refresh(self, project, providerURL=None, **kwargs):
        """Wraps `Rundeck API POST /project/[NAME]/resources/refresh <http://rundeck.org/docs/api/index.html#refreshing-resources-for-a-project>`_

        :Parameters:
            project : str
                name of the project
            providerURL : str
                Specify the Resource Model Provider URL to refresh the resources from; otherwise
                the configured provider URL in the `project.properties` file will be used

        :return: A :class:`~.rundeck.connection.RundeckResponse`
        :rtype: :class:`~.rundeck.connection.RundeckResponse`
        """
        self.requires_version(2)

        data = {}
        if providerURL is not None:
            data['providerURL'] = providerURL

        return self._exec(POST, 'project/{0}/resources/refresh'.format(project), data=data, **kwargs)

    def project_archive_import(self, project, archive_zip_data, **kwargs):
      """Wraps `Rundeck API PUT /api/14/project/[PROJECT]/import <http://rundeck.org/docs/api/index.html#project-archive-import>`_

      :Parameters:
          project : str
              name of the project
          archive_zip_data : ?
              Zip file containing a project archive.

      :Keywords:
          jobUuidOption : str ('preserve'|'remove')
              Option declaring how duplicate Job UUIDs should be handled. If preserve (default) then
              imported job UUIDs will not be modified, and may conflict with jobs in other projects.
              If remove then all job UUIDs will be removed before importing.
          importExecutions : bool
               import all executions and logs from the archive (default: true)
          importConfig : bool
               import project configurationfrom the archive (default: true)
          importACL : bool
               import all of the ACL policies from the archive (default: true)

      :return: A :class:`~.rundeck.connection.RundeckResponse`
      :rtype: :class:`~.rundeck.connection.RundeckResponse`
      """
      data = cull_kwargs(('jobUuidOption', 'importExecutions', 'importConfig', 'importACL'), kwargs)

      headers = {'Content-Type': 'application/zip'}

      return self._exec(PUT, 'project/{0}/import'.format(urlquote(project)), params=data, data=archive_zip_data, headers=headers, **kwargs)


    def history(self, project, **kwargs):
        """Wraps `Rundeck API GET /history <http://rundeck.org/docs/api/index.html#listing-history>`_

        :Parameters:
            project : str
                name of the project

        :Keywords:
            jobIdFilter : str
                include event for a job ID
            reportIdFilter : str
                include events for an event name
            userFilter : str
                include events created by user
            statFilter : str
                one of :class:`~rundeck.defaults.Status` ``values``
            jobListFilter : str | list
                one or more full job group/name to include
            excludeJobListFilter : str | list
                one or more full job group/name to include
            recentFilter | str
                Use a simple text format to filter executions that completed within a period of
                time; the format is "XY" where 'X' is an integer and 'Y' is one of:

                    * `h`:hour
                    * `d`:day
                    * `w`:week
                    * `m`:month
                    * `y`:year

                So a value of "2w" would return executions that completed within the last two weeks

            begin : int | str
                either a unix millisecond timestamp or a W3C dateTime "yyyy-MM-ddTHH:mm:ssZ"
            end : int | str
                either a unix millisecond timestamp or a W3C dateTime "yyyy-MM-ddTHH:mm:ssZ"
            max : int
                maximum number of results to include in response (default: 20)
            offset : int
                offset for result set (default: 0)


        :return: A :class:`~.rundeck.connection.RundeckResponse`
        :rtype: :class:`~.rundeck.connection.RundeckResponse`
        """
        self.requires_version(4)
        params = cull_kwargs(('jobIdFilter', 'reportIdFilter', 'userFilter', 'statFilter', \
            'jobListFilter', 'excludeJobListFilter', 'recentFilter', 'begin', 'end', 'max', \
            'offset'), kwargs)
        params['project'] = project
        return self._exec(GET, 'history', params=params, **kwargs)


class RundeckApi(RundeckApiTolerant):
    """ Same as RundeckApiTolerant and complains on every Rundeck Server error or 4xx/5xx HTTP
    status code.
    """
    def _exec(self, method, url, params=None, data=None, parse_response=True, **kwargs):
        quiet = kwargs.get('quiet', False)

        result = super(RundeckApi, self)._exec(
            method, url, params=params, data=data, parse_response=parse_response, **kwargs)

        if not quiet and parse_response:
            result.raise_for_error()

        return result



"""
:summary: Testing setup

:license: Apache License, Version 2.0
:author: Mark LaPerriere
:contact: otupman@antillion.com
:copyright: Mark LaPerriere 2015
"""
__docformat__ = "restructuredtext en"

import os
import logging
import uuid

from rundeck.client import Rundeck
from rundeck.api import RundeckApi, RundeckNode
from rundeck.transforms import _transforms as transforms

_RUNDECK_API_TOKEN_VAR = 'RUNDECK_API_TOKEN'
_RUNDECK_SERVER_VAR = 'RUNDECK_SERVER'
_RUNDECK_PORT_VAR = 'RUNDECK_PORT'
_RUNDECK_PROTOCOL_VAR = 'RUNDECK_PROTOCOL'
_RUNDECK_USR_VAR = 'RUNDECK_USR'
_RUNDECK_PWD_VAR = 'RUNDECK_PWD'
_RUNDECK_API_VERSION_VAR = 'RUNDECK_API_VERSION'

logger = logging.getLogger(__name__)
# logger = logging.getLogger('')  # for debugging
logger.setLevel(logging.INFO)
logger.addHandler(logging.StreamHandler())

cwd = os.path.dirname(os.path.realpath(__file__))
job_def_file = 'TestJobDef.xml'
job_def_path = os.path.join(cwd, job_def_file)

api_token = os.environ.get(_RUNDECK_API_TOKEN_VAR, None)
server = os.environ.get(_RUNDECK_SERVER_VAR, 'localhost')
port = os.environ.get(_RUNDECK_PORT_VAR, 4440)
protocol = os.environ.get(_RUNDECK_PROTOCOL_VAR, 'http')
usr = os.environ.get(_RUNDECK_USR_VAR, None)
pwd = os.environ.get(_RUNDECK_PWD_VAR, None)
api_version = os.environ.get(_RUNDECK_API_VERSION_VAR, 14)

test_job_id = uuid.uuid4()
test_job_name = 'TestJobTest'
test_job_proj = 'TestProjectTest'
test_job_def_tmpl = """<joblist>
  <job>
    <id>{0}</id>
    <loglevel>INFO</loglevel>
    <sequence keepgoing='false' strategy='node-first'>
      <command>
        <node-step-plugin type='localexec'>
          <configuration>
            <entry key='command' value='echo "Hello World! (from:${{option.from}})"' />
          </configuration>
        </node-step-plugin>
      </command>
      <command>
        <node-step-plugin type='localexec'>
          <configuration>
            <entry key='command' value='sleep ${{option.sleep}}' />
          </configuration>
        </node-step-plugin>
      </command>
    </sequence>
    <description></description>
    <name>{1}</name>
    <context>
      <project>{2}</project>
      <options>
        <option name='from' value='Tester' />
        <option name='sleep' value='0' />
      </options>
    </context>
    <uuid>{0}</uuid>
  </job>
</joblist>"""
test_job_def = test_job_def_tmpl.format(test_job_id, test_job_name, test_job_proj)

assert any([api_token, all([usr, pwd])]), 'Must specify a Rundeck API token or username/' + \
        'password pair with the {0} environment variable or use'.format(_RUNDECK_API_TOKEN_VAR) + \
        ' both the {0}, and {1}  environment variables'.format(_RUNDECK_USR_VAR, _RUNDECK_PWD_VAR)

options = {
    'server': server,
    'protocol': protocol,
    'port': port,
    'verify_cert': False,
    'api_version': api_version,
}

if api_token is not None:
    options['api_token'] = api_token
else:
    if usr is not None and pwd is not None:
        options['usr'] = usr
        options['pwd'] = pwd
    else:
        assert False, 'Must specifiy username AND password with the {0} and {1} ' + \
            'environment variables'.format(_RUNDECK_USR_VAR, _RUNDECK_PWD_VAR)

rundeck_client = Rundeck(**options)
rundeck_api = RundeckApi(**options)


def setup():
    rundeck_api.jobs_import(test_job_def, uuidOption='preserve')


def teardown():
    jobs = transforms['jobs'](rundeck_api.jobs(test_job_proj))
    rundeck_api.jobs_delete([job['id'] for job in jobs])



"""
:summary: Test rundeck.api.RundeckApi

:license: Apache License, Version 2.0
:author: Mark LaPerriere
:contact: otupman@antillion.com
:copyright: Mark LaPerriere 2015

:requires: requests
"""
__docformat__ = "restructuredtext en"

import uuid
import time
import requests
from datetime import datetime
from pprint import pprint, pformat

from nose.tools import raises

from tests import (
    rundeck_client,
    rundeck_api,
    test_job_id,
    test_job_name,
    test_job_proj,
    test_job_def_tmpl,
    transforms,
    RundeckNode,
    )

del_test_job_id = str(uuid.uuid4())
bulk1_del_test_job_id = str(uuid.uuid4())
bulk2_del_test_job_id = str(uuid.uuid4())
imp_test_job_id = str(uuid.uuid4())


def setup():
    rundeck_api.jobs_import(
        test_job_def_tmpl.format(
            del_test_job_id, test_job_name + 'ApiDelete', test_job_proj))
    rundeck_api.jobs_import(
        test_job_def_tmpl.format(
            bulk1_del_test_job_id, test_job_name + 'ApiBulkDelete1', test_job_proj))
    rundeck_api.jobs_import(
        test_job_def_tmpl.format(
            bulk2_del_test_job_id, test_job_name + 'ApiBulkDelete2', test_job_proj))
    time.sleep(3)


def teardown():
    pass


@raises(NotImplementedError)
def test_node_xml():
    # xml(self)
    raise NotImplementedError('Test not yet implemented')


@raises(NotImplementedError)
def test_exec():
    time.sleep(1)
    # _exec(self, method, url, params=None, data=None, parse_response=True, **kwargs)
    raise NotImplementedError('Test not yet implemented')


def test_system_info():
    time.sleep(1)
    assert rundeck_api.system_info().success, 'system_info call was unsuccessful'


def test_jobs():
    time.sleep(1)
    assert rundeck_api.jobs(test_job_proj).success, 'jobs call was unsuccessful'


def test_project_jobs():
    time.sleep(1)
    assert rundeck_api.project_jobs(test_job_proj).success, 'project_jobs call was unsuccessful'


def test_job_run():
    time.sleep(3)
    assert rundeck_api.job_run(test_job_id).success, 'job_run call was unsuccessful'
    time.sleep(3)
    assert rundeck_api.job_run(test_job_id, argString={'from': 'test_job_run'}).success, \
        'job_run with argument dict call was unsuccessful'
    time.sleep(3)
    assert rundeck_api.job_run(test_job_id, argString='-from test_job_run').success, \
        'job_run with argument string call was unsuccessful'


def test_jobs_export():
    time.sleep(1)
    assert rundeck_api.jobs_export(test_job_proj).status_code == requests.codes.ok, \
        'jobs_export call was unsuccessful'


@raises(NotImplementedError)
def test_jobs_import():
    time.sleep(1)
    # jobs_import(self, definition, **kwargs)
    raise NotImplementedError('Test not yet implemented')


def test_job():
    time.sleep(1)
    assert rundeck_api.job(test_job_id).status_code == requests.codes.ok, \
        'job call was unsuccessful'


def test_delete_job():
    time.sleep(1)
    response = rundeck_api.delete_job(del_test_job_id)
    assert response.status_code in (requests.codes.no_content, requests.codes.ok), \
        'delete_job call was unsuccessful'


def test_jobs_delete():
    time.sleep(1)
    assert rundeck_api.jobs_delete([bulk1_del_test_job_id, bulk2_del_test_job_id]).success, \
        'jobs_delete call was unsuccessful'


def test_job_executions():
    time.sleep(1)
    assert rundeck_api.job_executions(test_job_id).success, 'job_executions call was unsuccessful'


def test_executions_running():
    time.sleep(1)
    assert rundeck_api.executions_running(test_job_proj).success, \
        'executions_running call was unsuccessful'


def test_execution():
    time.sleep(1)
    execution = transforms['execution'](rundeck_api.job_run(test_job_id))
    time.sleep(1)
    execution_id = execution['id']
    assert rundeck_api.execution(execution_id).success, 'execution call was unsuccessful'


def test_executions():
    time.sleep(1)
    assert rundeck_api.executions(test_job_proj).success, 'executions call was unsuccessful'


def test_execution_output():
    time.sleep(1)
    execution = transforms['execution'](rundeck_api.job_run(test_job_id))
    time.sleep(1)
    execution_id = execution['id']
    assert rundeck_api.execution_output(execution_id).status_code == requests.codes.ok, \
        'execution_output call was unsuccessful'


def test_execution_abort():
    time.sleep(1)
    execution = transforms['execution'](rundeck_api.job_run(test_job_id, argString={'sleep': 30}))
    time.sleep(1)
    execution_id = execution['id']
    assert rundeck_api.execution_abort(execution_id).success, 'execution_abort call was unsuccessful'
    time.sleep(1)
    execution = transforms['execution'](rundeck_api.execution(execution_id))
    assert execution['status'] == 'aborted', 'execution_abort status is not \'aborted\''


@raises(NotImplementedError)
def test_run_command():
    time.sleep(1)
    # run_command(self, project, command, **kwargs)
    raise NotImplementedError('Test not yet implemented')


@raises(NotImplementedError)
def test_run_script():
    time.sleep(1)
    # run_script(self, project, scriptFile, **kwargs)
    raise NotImplementedError('Test not yet implemented')


@raises(NotImplementedError)
def test_run_url():
    time.sleep(1)
    # run_url(self, project, scriptURL, **kwargs)
    raise NotImplementedError('Test not yet implemented')


def test_projects():
    time.sleep(1)
    assert rundeck_api.projects().success, 'projects call was unsuccessful'


def test_project():
    time.sleep(1)
    assert rundeck_api.project(test_job_proj).success, 'project call was unsuccessful'


@raises(NotImplementedError)
def test_project_resources():
    time.sleep(1)
    # project_resources(self, project, **kwargs)
    raise NotImplementedError('Test not yet implemented')


def test_project_resources_update():
    time.sleep(1)
    host1 = RundeckNode(
        'host1', 'hostname1', 'user1', description='I <3 XML', tags=['hot'],
        attributes={'foo': 'bar'}
        )
    assert rundeck_api.project_resources_update(test_job_proj, [host1]).success, \
        'test_project_resources_update call was unsuccessful'


@raises(NotImplementedError)
def test_project_resources_refresh():
    time.sleep(1)
    # project_resources_refresh(self, project, providerURL=None, **kwargs)
    raise NotImplementedError('Test not yet implemented')


@raises(NotImplementedError)
def test_history():
    time.sleep(1)
    # history(self, project, **kwargs)
    raise NotImplementedError('Test not yet implemented')



"""
:summary: Default values

:license: Apache License, Version 2.0
:author: Mark LaPerriere
:contact: otupman@antillion.com
:copyright: Mark LaPerriere 2015
"""
__docformat__ = "restructuredtext en"


def enum(name, *sequential, **named):
    values = dict(zip(sequential, range(len(sequential))), **named)
    values['values'] = list(values.values())
    values['keys'] = list(values.keys())
    return type(name, (), values)

Status = enum(
    'Status',
    RUNNING='running',
    SUCCEEDED='succeeded',
    FAILED='failed',
    ABORTED='aborted',
    SKIPPED='skipped',
    PENDING='pending'
    )

DupeOption = enum(
    'DupeOption',
    SKIP='skip',
    CREATE='create',
    UPDATE='update'
    )

UuidOption = enum(
    'UuidOption',
    PRESERVE='preserve',
    REMOVE='remove'
    )

JobDefFormat = enum(
    'JobDefFormat',
    XML='xml',
    YAML='yaml'
    )

ExecutionOutputFormat = enum(
    'ExecutionOutputFormat',
    TEXT='text',
    **dict(zip(JobDefFormat.keys, JobDefFormat.values))
    )

RUNDECK_API_VERSION = 14
GET = 'get'
PUT = 'put'
POST = 'post'
DELETE = 'delete'
JOB_RUN_TIMEOUT = 60
JOB_RUN_INTERVAL = 3



__docformat__ = "restructuredtext en"
"""
:summary: Setup script for arundeckrun

:license: Apache License, Version 2.0
:author: Mark LaPerriere
:contact: otupman@antillion.com
:copyright: Mark LaPerriere 2015
"""
import os
from setuptools import setup, find_packages
from rundeck import VERSION

project = 'brundeckrun'
install_requires = ['requests>=2.23.0']

with open(os.path.join(os.path.dirname(os.path.realpath(__file__)), 'README.md'), 'r') as fp:
    long_description = fp.read()

setup(
    name=project,
    license='http://creativecommons.org/licenses/by-sa/3.0/',
    version=VERSION,
    packages=find_packages(exclude=["tests", "tests.*"]),
    
    description='A Rundeck API Python client',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/antillion/{0}'.format(project),
    author='rundeckrun@mindmind.com',
    author_email='rundeckrun@mindmind.com',

    maintainer='Lambda Vector',
    maintainer_email='pablo@ceruleanhq.com',

    install_requires=install_requires,

    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'Intended Audience :: System Administrators',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: System :: Software Distribution',
        'Topic :: System :: Systems Administration',
        'Topic :: Utilities',
    ],
)
